/*
	FDecimate plugin for Avisynth -- performs arbitrary
	decimation of a stream of progressive frames.

	Copyright (C) 2008 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Donald Graft
	neuron2@comcast.net.
*/

#include "internal.h"
#include "version.h"
#include "FDecimate.h"
#include "info.h"

PVideoFrame __stdcall FDecimate::GetFrame(int inframe, IScriptEnvironment* env)
{
	PVideoFrame src;
	char buf[255];
	double metric;
	int blind, choose, use;

	if (metrics == true)
	{
		// Don't decimate and show metrics for every frame.
		metric = CalculateMetrics(inframe - 1, inframe, env);
		GETFRAME(inframe, src);
	}
	else
	{
		// Decimate to achieve configured frame rate.
		// Avoiding delivering a duplicate if possible.
        blind = dec_data[inframe];
		if (last < 0)
			choose = blind;
		else if (CalculateMetrics(last, blind, env) > threshold)
			choose = blind;
		else if (CalculateMetrics(last, blind + 1, env) > threshold)
			choose = blind + 1;
		else if (CalculateMetrics(last, blind + 2, env) > threshold)
			choose = blind + 2;
        else
			choose = blind;
        use = choose;
		if (protect && CalculateMetrics(use, use + 1, env) <= threshold)
			use = use + 1;
 		GETFRAME(use, src);
		last = use;
	}

	// Debug displays.
	if (show == true || metrics == true)
	{
		env->MakeWritable(&src);
		if (vi.IsYUY2())
		{
			sprintf(buf, "FDecimate %s", VERSION);
			DrawStringYUY2(src, 0, 0, buf);
			sprintf(buf, "Copyright 2008 Donald Graft");
			DrawStringYUY2(src, 0, 1, buf);
			if (metrics == true)
				sprintf(buf,"in frm %d, metric %f", inframe, metric);
			else
				sprintf(buf,"in frm %d, blind frm %d, choose frm %d, use frm %d", inframe, blind, choose, use);
			DrawStringYUY2(src, 0, 3, buf);
		}
		else
		{
			sprintf(buf, "FDecimate %s", VERSION);
			DrawStringYV12(src, 0, 0, buf);
			sprintf(buf, "Copyright 2008 Donald Graft");
			DrawStringYV12(src, 0, 1, buf);
			if (metrics == true)
				sprintf(buf,"in frm %d, metric %f", inframe, metric);
			else
				sprintf(buf,"in frm %d, blind frm %d, choose frm %d, use frm %d", inframe, blind, choose, use);
			DrawStringYV12(src, 0, 3, buf);
		}
	}
	if (debug == true || metrics == true)
	{	
		if (metrics == true)
			sprintf(buf,"FDecimate: in frm %d, metric %f\n", inframe, metric);
		else
			sprintf(buf,"FDecimate: in frm %d, blind frm %d, choose frm %d, use frm %d", inframe, blind, choose, use);
		OutputDebugString(buf);
	}

	return src;
}

double FDecimate::CalculateMetrics(int a, int b, IScriptEnvironment* env)
{
	PVideoFrame prev, curr;
	const unsigned char *prevp, *currp, *prevpU, *currpU, *prevpV, *currpV;
	int currpitch, currpitchUV;
	int prevpitch, prevpitchUV;
	int row_size, row_sizeUV;
	int height, heightUV;
	int x, y, div;
	double metric;
	int xblocks, yblocks;
	unsigned int highest_sum;

	// Calculate a windowed sum-of-absolute-differences metric for the two
	// frames passed as parameters.
	GETFRAME(a, prev);
	GETFRAME(b, curr);

    currpitch = curr->GetPitch();
    prevpitch = prev->GetPitch();
    row_size = curr->GetRowSize();
    height = curr->GetHeight();
	if (vi.IsYV12())
	{
		currpitchUV = curr->GetPitch(PLANAR_U);
		prevpitchUV = prev->GetPitch(PLANAR_U);
		row_sizeUV = curr->GetRowSize(PLANAR_U);
		heightUV = curr->GetHeight(PLANAR_U);
	}
	div = (BLKSIZE * BLKSIZE / 3) * 255;

	xblocks = row_size / BLKSIZE;
	if (row_size % BLKSIZE) xblocks++;
	yblocks = height / BLKSIZE;
	if (height % BLKSIZE) yblocks++;

	if (sum == NULL)
	{
		sum = (unsigned int *) malloc(xblocks * yblocks * sizeof(unsigned int));
		if (sum == NULL)
			env->ThrowError("FDecimate: cannot allocate needed memory");
	}

	for (y = 0; y < yblocks; y++)
	{
		for (x = 0; x < xblocks; x++)
		{
			sum[y*xblocks+x] = 0;
		}
	}

	prevp = prev->GetReadPtr();
	currp = curr->GetReadPtr();
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < row_size;)
		{
			sum[(y/BLKSIZE)*xblocks+x/BLKSIZE] += abs((int)currp[x] - (int)prevp[x]);
			x++;
			if (!(x % 4)) x += 8;
		}
		prevp += prevpitch;
		currp += currpitch;
	}
	if (vi.IsYV12())
	{
		prevpU = prev->GetReadPtr(PLANAR_U);
		currpU = curr->GetReadPtr(PLANAR_U);
		prevpV = prev->GetReadPtr(PLANAR_V);
		currpV = curr->GetReadPtr(PLANAR_V);
		for (y = 0; y < heightUV; y++)
		{
			for (x = 0; x < row_sizeUV;)
			{
				sum[((2*y)/BLKSIZE)*xblocks+(2*x)/BLKSIZE] += abs((int)currpU[x] - (int)prevpU[x]);
				sum[((2*y)/BLKSIZE)*xblocks+(2*x)/BLKSIZE] += abs((int)currpV[x] - (int)prevpV[x]);
				x++;
				if (!(x % 4)) x += 8;
			}
			prevpU += prevpitchUV;
			currpU += currpitchUV;
			prevpV += prevpitchUV;
			currpV += currpitchUV;
		}
	}
	highest_sum = 0;
	for (y = 0; y < yblocks; y++)
	{
		for (x = 0; x < xblocks; x++)
		{
			if (sum[y*xblocks+x] > highest_sum)
			{
				highest_sum = sum[y*xblocks+x];
			}
		}
	}
	metric = (highest_sum * 100.0) / div;
	return metric;
}

// Borrowed from the author of IT.dll, whose name I
// could not determine. Modified for YV12 by Donald Graft.

void FDecimate::DrawDigitYUY2(PVideoFrame &dst, int x, int y, int num) 
{
	extern unsigned short font[][20];

	x = x * 10;
	y = y * 20;

	int pitch = dst->GetPitch();
	for (int tx = 0; tx < 10; tx++) {
		for (int ty = 0; ty < 20; ty++) {
			unsigned char *dp = &dst->GetWritePtr()[(x + tx) * 2 + (y + ty) * pitch];
			if (font[num][ty] & (1 << (15 - tx))) {
				if (tx & 1) {
					dp[0] = 250;
					dp[-1] = 128;
					dp[1] = 128;
				} else {
					dp[0] = 250;
					dp[1] = 128;
					dp[3] = 128;
				}
			} else {
				if (tx & 1) {
					dp[0] = (unsigned char) ((dp[0] * 3) >> 2);
					dp[-1] = (unsigned char) ((dp[-1] + 128) >> 1);
					dp[1] = (unsigned char) ((dp[1] + 128) >> 1);
				} else {
					dp[0] = (unsigned char) ((dp[0] * 3) >> 2);
					dp[1] = (unsigned char) ((dp[1] + 128) >> 1);
					dp[3] = (unsigned char) ((dp[3] + 128) >> 1);
				}
			}
		}
	}
}

void FDecimate::DrawStringYUY2(PVideoFrame &dst, int x, int y, const char *s) 
{
	for (int xx = 0; *s; ++s, ++xx) {
		DrawDigitYUY2(dst, x + xx, y, *s - ' ');
	}
}

void FDecimate::DrawDigitYV12(PVideoFrame &dst, int x, int y, int num) 
{
	int tx, ty;
	unsigned char *dpY, *dpU, *dpV;

	x = x * 10;
	y = y * 20;

	int pitchY = dst->GetPitch(PLANAR_Y);
	for (tx = 0; tx < 10; tx++)
	{
		for (ty = 0; ty < 20; ty++)
		{
			dpY = &dst->GetWritePtr(PLANAR_Y)[(x + tx) + (y + ty) * pitchY];
			if (font[num][ty] & (1 << (15 - tx)))
			{
				*dpY = 235;
			} else
			{
				*dpY = (unsigned char) ((*dpY * 3) >> 2);
			}
		}
	}
	int pitchUV = dst->GetPitch(PLANAR_V);
	for (tx = 0; tx < 10; tx++)
	{
		for (ty = 0; ty < 20; ty++)
		{
			dpU = &dst->GetWritePtr(PLANAR_U)[((x + tx)/2) + ((y + ty)/2) * pitchUV];
			dpV = &dst->GetWritePtr(PLANAR_V)[((x + tx)/2) + ((y + ty)/2) * pitchUV];
			if (font[num][ty] & (1 << (15 - tx)))
			{
				*dpU = 128;
				*dpV = 128;
			} else
			{
				*dpU = (unsigned char) ((*dpU + 128) >> 1);
				*dpV = (unsigned char) ((*dpV + 128) >> 1);
			}
		}
	}
}

void FDecimate::DrawStringYV12(PVideoFrame &dst, int x, int y, const char *s) 
{
	for (int xx = 0; *s; ++s, ++xx) {
		DrawDigitYV12(dst, x + xx, y, *s - ' ');
	}
}

AVSValue __cdecl Create_FDecimate(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char path[1024];
	char buf[80], *p;
	float rate = (float) 23.976;
	float threshold = 1.0;
	bool protect = false;
	bool metrics = false;
	bool show = false;
	bool debug = false;

	try
	{
		FILE *f;

		const char* plugin_dir = env->GetVar("$PluginDir$").AsString();
		strcpy(path, plugin_dir);
		strcat(path, "\\FDecimate.def");
		if ((f = fopen(path, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				if (strncmp(buf, "rate=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					rate = atof(p);
				}
				if (strncmp(buf, "threshold=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					threshold = atof(p);
				}
				if (strncmp(buf, "protect=", 8) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') protect = true;
					else protect = false;
				}
				if (strncmp(buf, "metrics=", 8) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') metrics = true;
					else metrics = false;
				}
				if (strncmp(buf, "show=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') show = true;
					else show = false;
				}
				if (strncmp(buf, "debug=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') debug = true;
					else debug = false;
				}
			}
			fclose(f);
		}
	}
	catch (...)
	{
		// plugin directory not set
	}

    return new FDecimate(args[0].AsClip(),		// clip
						args[1].AsFloat(rate),		// output frame rate
						args[2].AsFloat(threshold),	// duplicate threshold
						args[3].AsBool(protect),	// enable protection
						args[4].AsBool(metrics),	// show metrics mode
						args[5].AsBool(show),		// show
						args[6].AsBool(debug),		// debug
						env);
}

AVSValue __cdecl 
Create_FDecimate(AVSValue args, void* user_data, IScriptEnvironment* env);

extern "C" __declspec(dllexport) const char* __stdcall 
AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("FDecimate", "c[rate]f[threshold]f[protect]b[metrics]b[show]b[debug]b", Create_FDecimate, 0);
	return 0;
}
