/*
	FDecimate plugin for Avisynth -- performs arbitrary
	decimation of a stream of progressive frames.

	Copyright (C) 2008 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Donald Graft
	neuron2@comcast.net.
*/

#include "internal.h"
#include "version.h"

#define BLKSIZE 32

#define GETFRAME(g, fp) \
{ \
	int GETFRAMEf; \
	GETFRAMEf = (g); \
	if (GETFRAMEf < 0) GETFRAMEf = 0; \
	if (GETFRAMEf > num_frames_hi - 1) GETFRAMEf = num_frames_hi - 1; \
	(fp) = child->GetFrame(GETFRAMEf, env); \
}

class FDecimate : public GenericVideoFilter
{
private:
	float rate, threshold;
	bool protect, metrics, debug, show;
	int num_frames_hi, num_frames_lo;
	unsigned int *dec_data;
	int last;
	unsigned int *sum;

    PVideoFrame __stdcall FDecimate::GetFrame(int n, IScriptEnvironment* env);
    double __stdcall FDecimate::CalculateMetrics(int frame, int last, IScriptEnvironment* env);
	void FDecimate::DrawStringYUY2(PVideoFrame &dst, int x, int y, const char *s);
	void FDecimate::DrawDigitYUY2(PVideoFrame &dst, int x, int y, int num); 
	void FDecimate::DrawDigitYV12(PVideoFrame &dst, int x, int y, int num); 
	void FDecimate::DrawStringYV12(PVideoFrame &dst, int x, int y, const char *s);
	
public:
	FDecimate::FDecimate(PClip _child, float _rate,
				float _threshold, bool _protect, bool _metrics, bool _show, bool _debug, IScriptEnvironment* env) :
		        GenericVideoFilter(_child), rate(_rate),
				threshold(_threshold), protect(_protect), metrics(_metrics), show(_show), debug(_debug)
	{
		int i;
		double n = rate;
		double fps = (double) vi.fps_numerator / vi.fps_denominator;
		int d = 1;

		if (!vi.IsYUY2() &&!vi.IsYV12() )
			env->ThrowError("FDecimate: input must be YUY2 or YV12");
		if (rate < 1)
			env->ThrowError("FDecimate: frame rate invalid");
		if (threshold < 0)
			env->ThrowError("FDecimate: threshold must be positive");

		/* Adjust frame rate and count. */
		dec_data = NULL;
		if (metrics == true)
		{
			// Just looking at metrics, so don't decimate.
			num_frames_hi = num_frames_lo = vi.num_frames;
		}
		else
		{
			while (n < 16777216 && d < 16777216)
			{
				n*=2;
				d*=2;
			}
			vi.SetFPS(int(n+0.5), d);
			num_frames_hi = vi.num_frames;
			num_frames_lo = (int) (num_frames_hi * (rate/fps));
			vi.num_frames = num_frames_lo;

			// Calculate the frames to deliver based on naive frame
			// selection. This can be modified at delivery time to
			// avoid delivering duplicates.
			dec_data = (unsigned int *) malloc(4 * num_frames_lo);
			if (dec_data == NULL)
				env->ThrowError("FDecimate: cannot allocate needed memory");
			for (i = 0; i < num_frames_lo; i++)
			{
				dec_data[i] = (int)(i * fps / rate);
			}
			last = -1;
		}

		sum = NULL;

		if (debug)
		{
			char b[80];
			sprintf(b, "FDecimate %s by Donald Graft, Copyright 2008\n", VERSION);
			OutputDebugString(b);
		}
	}

	FDecimate::~FDecimate()
	{
		if (dec_data) free(dec_data);
		if (sum) free(sum);
	}
};

